import 'dart:async';

abstract class Result<T> {
  T? val;
  Error? err;
  bool get isOk => this is Ok;
  bool get isErr => this is Err;
  T? get unwrap {
    if (isErr) {
      throw err!;
    }
    return val;
  }

  @override
  String toString()=>isErr ? 'Err($err)' : 'Ok($val)';

  T? unwrapOrDefault(T default_value) {
    if (isErr) {
      return default_value;
    }
    return val;
  }

  T? unwrapOrElse(T Function(Error? e) elseFunc) {
    if (isErr) {
      return elseFunc(err);
    }
    return val;
  }
}

class Ok<T> extends Result<T> {
  @override
  T? val;
  Ok(this.val) {
    ArgumentError.checkNotNull(val);
  }
}

class Err<T> extends Result<T> {
  @override
  Error? err;
  Err(this.err);
}

/// Calls `Function` f and awaits it.
/// Returns Ok(...) if all right
/// Returns Err(...) otherwise
Future<Result<T>> tryAsResultAsync<T>(FutureOr<T> Function() f) async {
  try {
    var callResult = await f();
    return callResult == null
        ? Err<T>(StateError('async function returned null'))
        : Ok(callResult);
  } catch (e) {
    return Err(StateError('$e'));
  }
}

/// Calls `Function` f
/// Returns Ok(...) if all right
/// Returns Err(...) otherwise
Result<T> tryAsResultSync<T>(T Function() f) {
  try {
    var callResult = f();
    return callResult == null
        ? Err<T>(StateError('function returned null'))
        : Ok(callResult);
  } catch (e) {
    return Err(StateError('$e'));
  }
}

/// Calls test(result) and returns result as is if test return true.
/// Otherwise returns Err(ArgumentError(failedTestMsg)).
Result<T> testResult<T>(bool? Function(Result<T>) test, Result<T> result,
        String failedTestMsg) =>
    test(result)! ? result : Err(ArgumentError(failedTestMsg));

/// If left-side `Result r` is a Err, then returns the same `Result r`.
/// Otherwise return f(r.val).
/// If result of function call `res` is Result - then returns it as is.
/// Otherwise if result of function call `res` == null - returns Err,
/// else returns Ok(res).
extension ResultPiping<T> on Result<T> {
  Result operator |(Function(T? x) f) {
    if (isErr) {
      return this;
    }
    var res = f(val);
    return res == null
        ? Err(StateError('function returned null'))
        : res is Result ? res : Ok(res);
  }
}
